from .bounds import Bounds
from .frame import Frame
from .map import Map, MapAlignment
from .orientation import Orientation
from .pose import Pose
from .position import Position, Positions
from .translation import Translation



from __future__ import annotations

from dataclasses import dataclass
from typing import Union

import numpy as np
from scipy.spatial.transform import Rotation

from .models.frame import Frame
from .models.orientation import Orientation
from .models.pose import Pose
from .models.position import Position, Positions
from .models.translation import Translation


@dataclass
class Transform:
    """
    A transform object that describe the transformation between two frames.
    Contains a scipy rotation object, a translation and two frames.
    Can be created from euler array or quaternion array. Translations must be
    expressed in the (to_) frame
    """

    translation: Translation
    from_: Frame
    to_: Frame
    rotation: Rotation = None

    def __post_init__(self):
        if (
            not self.translation.from_ == self.from_
            and not self.translation.to_ == self.to_
        ):
            raise ValueError(
                f"The from_ frames or to_ frames of translation and transform object are not equal."
            )

    def transform_position(
        self,
        positions: Union[Position, Positions],
        from_: Frame,
        to_: Frame,
    ) -> Union[Position, Positions]:
        """
        Transforms a position or list of positions from from_ to to_ (rotation and translation)
        :param positions: Position or Positions in the from_ coordinate system.
        :param from_: Source Frame, must be different to "to_".
        :param to_: Destination Frame, must be different to "from_".
        :return: Position or Positions in the to_ coordinate system.
        """
        if positions.frame != from_:
            raise ValueError(
                f"Expected positions in frame {from_} "
                + f", got positions in frame {positions.frame}"
            )

        if from_ == to_:
            return positions

        result: np.ndarray
        if from_ == self.to_ and to_ == self.from_:
            """Using the inverse transform"""
            result = self.rotation.apply(
                positions.to_array() - self.translation.to_array(),
                inverse=True,
            )
        elif from_ == self.from_ and to_ == self.to_:
            result = (
                self.rotation.apply(positions.to_array()) + self.translation.to_array()
            )
        else:
            raise ValueError("Transform not specified")

        if isinstance(positions, Position):
            return Position.from_array(result, to_)
        elif isinstance(positions, Positions):
            return Positions.from_array(result, to_)
        else:
            raise ValueError("Incorrect input format. Must be Position or Positions.")

    def transform_rotation(
        self, rotation: Rotation, from_: Frame, to_: Frame
    ) -> Rotation:
        """
        Transforms a rotation from from_ to to_ (rotation)
        :param rotation: Rotation (scipy) in the from_ coordinate system.
        :param from_: Source Frame, must be different to "to_".
        :param to_: Destination Frame, must be different to "from_".
        :return: Rotation in the to_ coordinate system.
        """

        if from_ == self.to_ and to_ == self.from_:
            "Using the inverse transform"
            rotation_to = rotation * self.rotation.inv()
        elif from_ == self.from_ and to_ == self.to_:
            rotation_to = rotation * self.rotation
        else:
            raise ValueError("Transform not specified")

        return rotation_to

    def transform_orientation(
        self,
        orientation: Orientation,
        from_: Frame,
        to_: Frame,
    ) -> Orientation:
        """
        Transforms an orientation from from_ to to_ (rotation)
        :param orientation: Orientation in the from_ coordinate system.
        :param from_: Source Frame, must be different to "to_".
        :param to_: Destination Frame, must be different to "from_".
        :return: Orientation in the to_ coordinate system.
        """

        if from_ == to_:
            return orientation

        rotation_to = self.transform_rotation(orientation.to_rotation(), from_, to_)
        return Orientation(*rotation_to.as_quat(), frame=to_)  # type: ignore

    def transform_pose(self, pose: Pose, from_: Frame, to_: Frame) -> Pose:
        """
        Transforms a pose from from_ to to_ (rotation)
        :param pose: Pose in the from_ coordinate system.
        :param from_: Source Frame, must be different to "to_".
        :param to_: Destination Frame, must be different to "from_".
        :return: Pose in the to_ coordinate system.
        """

        if from_ == to_:
            return pose

        position = self.transform_position(pose.position, from_, to_)
        if not isinstance(position, Position):
            raise TypeError("Pose can only contain a single position, not positions")
        orientation: Orientation = self.transform_orientation(
            pose.orientation, from_, to_
        )

        return Pose(position, orientation, to_)

    @staticmethod
    def from_euler_array(
        translation: Translation, euler: np.ndarray, from_: Frame, to_: Frame, seq="ZYX"
    ) -> Transform:
        """
        :param translation: Translation object between two frames
        :param euler: Numpy array of euler angles [x,y,z], shape (3,)
        :param from_: Frame the transform is coming from
        :param to_: Frame the transform is going to
        :param seq: Sequence of axes for rotations, same as scipy rotation
        :return: Transform object
        """
        return Transform(
            translation=translation,
            from_=from_,
            to_=to_,
            rotation=Rotation.from_euler(seq=seq, angles=euler),
        )

    @staticmethod
    def from_quat_array(
        translation: Translation,
        quat: np.ndarray,
        from_: Frame,
        to_: Frame,
    ) -> Transform:
        """
        :param translation: Translation object between two frames
        :param quat: Numpy array of quaternions [x,y,z,w], shape (4,)
        :param from_: Frame the transform is coming from
        :param to_: Frame the transform is going to
        :return: Transform object
        """
        rotation = Rotation.from_quat(quat)
        return Transform(
            translation=translation,
            from_=from_,
            to_=to_,
            rotation=rotation,
        )



from __future__ import annotations

from dataclasses import dataclass

import numpy as np

from alitra.models import Frame


@dataclass
class Translation:
    """
    A translation between two frames represented as x, y and z
    """

    x: float
    y: float
    from_: Frame
    to_: Frame
    z: float = 0

    def to_array(self) -> np.ndarray:
        """
        :return: Numpy array of translation, shape (3,)
        """
        return np.array([self.x, self.y, self.z], dtype=float)

    @staticmethod
    def from_array(
        translation: np.ndarray,
        from_: Frame,
        to_: Frame,
    ) -> Translation:
        """
        :param translation: Numpy array of translation [x,y,z].
            Needs to be shape (3,)
        :param from_: Frame the translation is coming from
        :param to_: Frame the translation is going to
        """
        if translation.shape != (3,):
            raise ValueError("Translation should have shape (3,)")
        return Translation(
            x=translation[0], y=translation[1], z=translation[2], from_=from_, to_=to_
        )



from dataclasses import dataclass

from .position import Position


@dataclass
class Bounds:
    """
    Bounds defines the cube bounds of which the map should be valid inside.
    Position 1 and 2 should be on the diagonal corners of the bounding cube.
    """

    position1: Position
    position2: Position

    def __post_init__(self):
        if self.position1.frame == self.position2.frame:
            self.frame = self.position1.frame
        else:
            raise TypeError("The frames of the bounding Positions are not the same")
        positions: list = [self.position1, self.position2]
        self.x_max = max(position.x for position in positions)
        self.x_min = min(position.x for position in positions)
        self.y_max = max(position.y for position in positions)
        self.y_min = min(position.y for position in positions)
        self.z_max = max(position.z for position in positions)
        self.z_min = min(position.z for position in positions)

    def position_within_bounds(self, position: Position) -> bool:
        if not position.frame == self.frame:
            raise ValueError(
                f"The position is in {position.frame} frame and the bounds are in {self.frame} frame"
            )
        if position.x < self.x_min or position.x > self.x_max:
            return False
        if position.y < self.y_min or position.y > self.y_max:
            return False
        if position.z < self.z_min or position.z > self.z_max:
            return False
        return True



from __future__ import annotations

import json
from dataclasses import dataclass
from pathlib import Path

from dacite import from_dict

from .bounds import Bounds
from .frame import Frame
from .position import Positions


@dataclass
class Map:
    """
    Map contains a set of coordinates in a given frame. Two maps can be used to create
    a transform. A map may also contain bounds.
    """

    name: str
    frame: Frame
    reference_positions: Positions
    bounds: Bounds = None

    @staticmethod
    def from_config(map_config_path: Path) -> Map:
        """
        Loads a Map from a json-file using dacite
        """
        with open(map_config_path) as json_file:
            map_config_dict = json.load(json_file)

        return from_dict(data_class=Map, data=map_config_dict)


@dataclass
class MapAlignment:
    """
    MapAlignment contains two maps that can be used to create a transform
    """

    name: str
    map_from: Map
    map_to: Map

    @staticmethod
    def from_config(map_config_path: Path) -> MapAlignment:
        """
        Loads a MapAlignment from a json-file using dacite
        """
        with open(map_config_path) as json_file:
            map_config_dict = json.load(json_file)

        return from_dict(data_class=MapAlignment, data=map_config_dict)



"""
Alitra makes alignment and transformation between coordinate-frames
easier by using custom dataclasses

Imagine we have an "asset" with one coordinate-frame and a "robot" with its own
internal coordinate-frame. We want to transform a position in the robot coordinate-frame
to a position in the asset coordinate-frame.

>>> import numpy as np
>>> from alitra.models import Positions, Transform, Translation

Setting up rotations, translations and positions for transformation

>>> robot_frame = Frame("robot")
>>> asset_frame = Frame("asset")
>>> euler = np.array([np.pi / 4, 0, 0])
>>> translation = Translation(x=1, y=0, from_=robot_frame, to_=asset_frame)
>>> p_robot = Positions.from_array(np.array([[1, 1, 0], [10, 1, 0]]), frame=robot_frame)
>>> rotation_axes = "z"

Making the transform

>>> transform = Transform.from_euler_array(
...    translation=translation, euler=euler, from_=robot_frame, to_=asset_frame
... )

Tranform position on robot to a position on the asset

>>> p_asset = transform.transform_position(p_robot, from_=robot_frame, to_=asset_frame)

If you have one position in two different frames and the rotation between the axes you can find the transform
between the two frames.

>>> transform = Transform(p_robot, p_asset, rotation_axes)
"""

from alitra.alignment import align_maps, align_positions
from alitra.models import (
    Bounds,
    Frame,
    Map,
    MapAlignment,
    Orientation,
    Pose,
    Position,
    Positions,
    Translation,
)
from alitra.transform import Transform



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class Frame:
    """
    Frame is used by most of our models to describe in which frame a model lives,
    or the two frames a transform is between.
    """

    name: str



from __future__ import annotations

from dataclasses import dataclass
from typing import Tuple

import numpy as np

from .frame import Frame
from .orientation import Orientation
from .position import Position


@dataclass
class Pose:
    """
    Pose contains a position, an orientation and a frame
    """

    position: Position
    orientation: Orientation
    frame: Frame

    def to_array(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        :return: Tuple of numpy arrays of the position and orientation as a quaternion
            with shapes (3,) and (4,) respectively
        """
        return self.position.to_array(), self.orientation.to_quat_array()

    @staticmethod
    def from_array(pos_array: np.ndarray, quat_array: np.ndarray, frame: Frame) -> Pose:
        """
        :param pos_array: Numpy array of shape (3,) containing position [x,y,z]
        :param quat_array: Numpy array of shape (4,) containing orientation as a
            quaternion [x,y,z,w]
        :param frame: Frame of pose
        :return: Pose object
        """
        return Pose(
            Position.from_array(pos_array, frame),
            Orientation.from_quat_array(quat_array, frame),
            frame,
        )



from __future__ import annotations

from dataclasses import dataclass
from typing import List

import numpy as np

from .frame import Frame


@dataclass
class Position:
    """
    Position contains the x, y and z coordinate as well as a frame
    """

    x: float
    y: float
    z: float
    frame: Frame

    def to_array(self) -> np.ndarray:
        """
        :return: Numpy array of position, shape (3,)
        """
        return np.array([self.x, self.y, self.z], dtype=float)

    @staticmethod
    def from_array(position: np.ndarray, frame: Frame) -> Position:
        """
        :param position: Numpy array of position [x,y,z].
            Needs to be shape (3,)
        :param frame: Frame of position
        """
        if position.shape != (3,):
            raise ValueError("position array must have shape (3,)")
        return Position(x=position[0], y=position[1], z=position[2], frame=frame)


@dataclass
class Positions:
    """
    Positions contains a list of positions as well as a frame in which the points
    are valid
    """

    positions: List[Position]
    frame: Frame

    def to_array(self) -> np.ndarray:
        """
        :return: Numpy array of positions, shape (N,3)
        """
        positions = []
        for position in self.positions:
            positions.append([position.x, position.y, position.z])
        return np.array(positions, dtype=float)

    @staticmethod
    def from_array(position_array: np.ndarray, frame: Frame) -> Positions:
        """
        :param position_array: Numpy array of positions i.e. [[x,y,z],[x,y,z]].
            Needs to be shape (N,3)
        :param frame: Frame of positions
        """
        if len(position_array.shape) < 2 or position_array.shape[1] != 3:
            raise ValueError("position_array should have shape (N,3)")
        positions: List[Position] = []
        for position in position_array:
            positions.append(
                Position(x=position[0], y=position[1], z=position[2], frame=frame)
            )
        return Positions(positions=positions, frame=frame)



from typing import Literal, Tuple

import numpy as np
from numpy.linalg import norm  # type: ignore
from scipy.spatial.transform import Rotation

from .models.map import Map
from .models.position import Positions
from .models.translation import Translation
from .transform import Transform


def align_maps(
    map_from: Map,
    map_to: Map,
    rot_axes: Literal["x", "y", "z", "xyz"],
    rsmd_threshold=0.4,
) -> Transform:
    """
    Uses align_positions to create a transform between two maps.
    See align_positions for further information.

    :param map_from: Map the transform is coming from
    :param map_to: Map the transform is going to
    :param rot_axes: Axis of rotation. For rotations in the xy plane (most common),
        this is set to 'z'
    :param rsmd_threshold: The root mean square distance threshold, for the coordinate
        fitting error in matching the two coordinate systems.
    """
    return align_positions(
        map_from.reference_positions,
        map_to.reference_positions,
        rot_axes,
        rsmd_threshold,
    )


def align_positions(
    positions_from: Positions,
    positions_to: Positions,
    rot_axes: Literal["x", "y", "z", "xyz"],
    rsmd_threshold=0.4,
) -> Transform:
    """
    Let positions_from be fixed local coordinate frame, and positions_to be some other
    fixed coordinate system. Further, let the relationship between the two reference
    systems be described as

    positions_from = rotation.apply(positions_to) + translation

    This function finds the rotation and translation by matching the two
    coordinate systems, and represent the transformation through a Transform object.
    For robustness it is advised to use more than 2 positions in alignment and using
    positions with some distance to each other.
    :param positions_from: Coordinates in a fixed frame
    :param positions_to: Coordinates in a fixed frame
    :param rot_axes: Axis of rotation. For rotations in the xy plane (most common),
        this is set to 'z'
    :param rsmd_threshold: The root mean square distance threshold, for the coordinate
        fitting error in matching the two coordinate systems.
    """
    if len(positions_from.positions) != len(positions_to.positions):
        raise ValueError(
            f"Expected inputs 'positions_from' and 'positions_to' to have the same shapes"
            + f" got {len(positions_from.positions)} and {len(positions_to.positions)}, respectively"
        )
    if len(positions_from.positions) < 2:
        raise ValueError(
            f" Expected at least 2 positions, got {len(positions_from.positions)}"
        )
    if len(positions_from.positions) < 3 and rot_axes == "xyz":
        raise ValueError(
            f" Expected at least 3 positions, got {len(positions_from.positions)}"
        )

    try:
        edges_1, edges_2 = _get_edges(positions_from, positions_to, rot_axes)
    except Exception as e:
        raise ValueError(e)

    rotation, rmsd_rot, sensitivity = Rotation.align_vectors(
        edges_2, edges_1, return_sensitivity=True
    )

    translations: Translation = Translation.from_array(
        np.mean(
            positions_to.to_array() - rotation.apply(positions_from.to_array()),
            axis=0,  # type:ignore
        ),
        from_=positions_from.frame,
        to_=positions_to.frame,
    )  # type: ignore
    transform = Transform(
        from_=positions_from.frame,
        to_=positions_to.frame,
        translation=translations,
        rotation=rotation,
    )

    try:
        _check_rsme_treshold(transform, positions_to, positions_from, rsmd_threshold)
    except Exception as e:
        raise ValueError(e)

    return transform


def _get_edges(
    positions_from: Positions,
    positions_to: Positions,
    rot_axes: Literal["x", "y", "z", "xyz"],
    tol: float = 10e-2,
) -> Tuple[np.ndarray, np.ndarray]:
    edges_from = _get_edges_between_coordinates(positions_from)
    edges_to = _get_edges_between_coordinates(positions_to)
    if np.min(norm((np.vstack([edges_from, edges_to])), axis=1)) < tol:
        raise ValueError("Positions are not unique")
    edges_from, edges_to = _add_dummy_rot_axis_edge(edges_from, edges_to, rot_axes)
    return edges_from, edges_to


def _get_edges_between_coordinates(positions_from: Positions) -> np.ndarray:
    """Finds all edges (vectors) between the input coordinates"""
    positions_from_arr = positions_from.to_array()
    n_positions = positions_from_arr.shape[0]
    edges = np.empty([sum(range(n_positions)), 3])
    index = 0
    for i in range(0, n_positions - 1):
        for j in range(i + 1, n_positions):
            edges[index, :] = np.array(
                [
                    positions_from_arr[i] - positions_from_arr[j],
                ]
            )
            index = index + 1
    return edges


def _add_dummy_rot_axis_edge(
    edges_from: np.ndarray,
    edges_to: np.ndarray,
    rot_axes: Literal["x", "y", "z", "xyz"],
) -> Tuple[np.ndarray, np.ndarray]:
    """Adds vectors to help ensure no rotations about non specified axes"""
    if rot_axes == "z":
        edges_from[:, 2] = 0
        edges_to[:, 2] = 0
        edges_from = np.vstack([edges_from, np.array([[0, 0, 1]])])
        edges_to = np.vstack([edges_to, np.array([[0, 0, 1]])])
    if rot_axes == "y":
        edges_from[:, 1] = 0
        edges_to[:, 1] = 0
        edges_from = np.vstack([edges_from, np.array([[0, 1, 0]])])
        edges_to = np.vstack([edges_to, np.array([[0, 1, 0]])])
    if rot_axes == "x":
        edges_from[:, 0] = 0
        edges_to[:, 0] = 0
        edges_from = np.vstack([edges_from, np.array([[1, 0, 0]])])
        edges_to = np.vstack([edges_to, np.array([[1, 0, 0]])])
    return edges_from, edges_to


def _check_rsme_treshold(
    transform: Transform,
    positions_to: Positions,
    positions_from: Positions,
    rsmd_threshold: float,
) -> float:
    positions_to_new = transform.transform_position(
        positions_to, from_=positions_to.frame, to_=positions_from.frame
    )
    transform_distance_error = positions_from.to_array() - positions_to_new.to_array()
    rsm_distance = np.mean(norm(transform_distance_error, axis=1))
    if rsm_distance > rsmd_threshold:
        raise ValueError(
            f"Root mean square error {rsm_distance:.4f} exceeds treshold {rsmd_threshold}"
        )
    return rsmd_threshold



from __future__ import annotations

from dataclasses import dataclass

import numpy as np
from scipy.spatial.transform import Rotation

from .frame import Frame


@dataclass
class Orientation:
    """
    This class represents an orientation using the quaternion values:
    x, y, z, w, and a frame. This classes uses scipy rotation to handle transformations
    between euler and quaternions
    """

    x: float
    y: float
    z: float
    w: float
    frame: Frame

    def to_euler_array(
        self, degrees: bool = False, wrap_angles: bool = False, seq: str = "ZYX"
    ) -> np.ndarray:
        """
        :param degrees: Set to true to retrieve angles as degrees
        :param wrap_angles: Set to true to get angles between 0 and 360 deg or
            0 and two pi
        :param seq: Sequence of axes for rotations, same as scipy rotation
        :return: Numpy array of euler angles
        """
        rotation: Rotation = Rotation.from_quat(self.to_quat_array())
        euler = rotation.as_euler(seq=seq, degrees=degrees)

        if wrap_angles:
            base = 360.0 if degrees else 2 * np.pi
            euler = np.fromiter(map(lambda angle: (angle % base), euler), dtype=float)

        return euler

    def to_quat_array(self) -> np.ndarray:
        """
        :return: Numpy array of quaternion values, [x,y,z,w]
        """
        return np.array([self.x, self.y, self.z, self.w], dtype=float)

    def to_rotation(self) -> Rotation:
        """
        :return: Scipy Rotation object
        """
        return Rotation.from_quat(self.to_quat_array())

    @staticmethod
    def from_quat_array(quat: np.ndarray, frame: Frame) -> Orientation:
        """
        :param quat: Numpy array of shape (4,) containing quaternion values, [x,y,z,w]
        :param frame: Frame of orientation
        :return: Orientation object
        """
        if quat.shape != (4,):
            raise ValueError("quaternion should have shape (4,)")
        return Orientation(
            *quat,  # type: ignore
            frame=frame,
        )

    @staticmethod
    def from_euler_array(
        euler: np.ndarray, frame: Frame, degrees: bool = False, seq: str = "ZYX"
    ) -> Orientation:
        """
        :param euler: Numpy array of euler angles of shape (3,)
        :param frame: Frame of orientation
        :param degrees: Set to true to retrieve angles as degrees
        :param wrap_angles: Set to true to get angles between 0 and 360 deg or
            0 and two pi
        :param seq: Sequence of axes for rotations, same as scipy rotation
        :return: Orientation object
        """
        rotation = Rotation.from_euler(seq=seq, angles=euler, degrees=degrees)
        return Orientation(*rotation.as_quat(), frame=frame)  # type: ignore

    @staticmethod
    def from_rotation(rotation: Rotation, frame: Frame) -> Orientation:
        """
        :param rotation: Scipy Rotation object
        :param frame: Frame of orientation
        :return: Orientation object
        """
        return Orientation(*rotation.as_quat(), frame=frame)  # type: ignore



from setuptools import find_packages, setup

setup(
    name="alitra",
    description="Simple alignment and transformation between coordinate frames",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    version="1.1.0",
    author="Equinor ASA",
    author_email="arnts@equinor.com, euel@equinor.com, chjo@equinor.com",
    license="MIT",
    url="https://github.com/equinor/alitra",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    classifiers=[
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Software Development :: Libraries",
    ],
    include_package_data=True,
    install_requires=["scipy", "numpy", "dacite"],
    python_requires=">=3.8",
    extras_require={"dev": ["pytest", "black", "mypy"]},
)
